library expanding_picker_sheet;

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

const Cubic _kAccelerateCurve = Cubic(0.548, 0.0, 0.757, 0.464);
const Cubic _kDecelerateCurve = Cubic(0.23, 0.94, 0.41, 1.0);
const double _kPeakVelocityTime = 0.248210;
const double _kPeakVelocityProgress = 0.379146;
const double _kCartHeight = 56.0;
const double _kCornerRadius = 24.0;
const double _kWidthForCartIcon = 64.0;

class ExpandingPickerSheet extends StatefulWidget {
  final AnimationController hideController;
  final Widget sheetThumbnail;
  final Widget pickerScreen;

  const ExpandingPickerSheet({
    Key key,
    @required this.hideController,
    this.sheetThumbnail,
    this.pickerScreen,
  })  : assert(hideController != null),
        super(key: key);

  @override
  _ExpandingPickerSheetState createState() => _ExpandingPickerSheetState();

  static _ExpandingPickerSheetState of(BuildContext context,
      {bool isNullOk = false}) {
    assert(isNullOk != null);
    assert(context != null);
    final _ExpandingPickerSheetState result = context.ancestorStateOfType(
            const TypeMatcher<_ExpandingPickerSheetState>())
        as _ExpandingPickerSheetState;
    if (isNullOk || result != null) {
      return result;
    }
    throw FlutterError(
        'ExpandingPickerSheet.of() called with a context that does not contain a ExpandingPickerSheet.\n');
  }
}

Animation<T> _getEmphasizedEasingAnimation<T>(
    {@required T begin,
    @required T peak,
    @required T end,
    @required bool isForward,
    @required Animation<double> parent}) {
  Curve firstCurve;
  Curve secondCurve;
  double firstWeight;
  double secondWeight;

  if (isForward) {
    firstCurve = _kAccelerateCurve;
    secondCurve = _kDecelerateCurve;
    firstWeight = _kPeakVelocityTime;
    secondWeight = 1.0 - _kPeakVelocityTime;
  } else {
    firstCurve = _kDecelerateCurve.flipped;
    secondCurve = _kAccelerateCurve.flipped;
    firstWeight = 1.0 - _kPeakVelocityTime;
    secondWeight = _kPeakVelocityTime;
  }

  return TweenSequence(
    <TweenSequenceItem<T>>[
      TweenSequenceItem<T>(
        weight: firstWeight,
        tween: Tween<T>(
          begin: begin,
          end: peak,
        ).chain(CurveTween(curve: firstCurve)),
      ),
      TweenSequenceItem<T>(
        weight: secondWeight,
        tween: Tween<T>(
          begin: peak,
          end: end,
        ).chain(CurveTween(curve: secondCurve)),
      ),
    ],
  ).animate(parent);
}

double _getPeakPoint({double begin, double end}) {
  return begin + (end - begin) * _kPeakVelocityProgress;
}

class _ExpandingPickerSheetState extends State<ExpandingPickerSheet>
    with TickerProviderStateMixin {
  final GlobalKey _expandingPickerSheetKey =
      GlobalKey(debugLabel: 'Expanding bottom sheet');

  double _width = _kWidthForCartIcon;

  AnimationController _controller;

  Animation<double> _widthAnimation;
  Animation<double> _heightAnimation;
  Animation<double> _thumbnailOpacityAnimation;
  Animation<double> _cartOpacityAnimation;
  Animation<double> _shapeAnimation;
  Animation<Offset> _slideAnimation;

  @override
  void initState() {
    super.initState();
    _controller = AnimationController(
      duration: const Duration(milliseconds: 500),
      vsync: this,
    );
  }

  @override
  void dispose() {
    _controller.dispose();
    super.dispose();
  }

  Animation<double> _getWidthAnimation(double screenWidth) {
    if (_controller.status == AnimationStatus.forward) {
      return Tween<double>(begin: _width, end: screenWidth).animate(
        CurvedAnimation(
          parent: _controller.view,
          curve: Interval(0.0, 0.3, curve: Curves.fastOutSlowIn),
        ),
      );
    } else {
      return _getEmphasizedEasingAnimation(
        begin: _width,
        peak: _getPeakPoint(begin: _width, end: screenWidth),
        end: screenWidth,
        isForward: false,
        parent: CurvedAnimation(
            parent: _controller.view, curve: Interval(0.0, 0.87)),
      );
    }
  }

  Animation<double> _getHeightAnimation(double screenHeight) {
    if (_controller.status == AnimationStatus.forward) {
      return _getEmphasizedEasingAnimation(
        begin: _kCartHeight,
        peak: _kCartHeight +
            (screenHeight - _kCartHeight) * _kPeakVelocityProgress,
        end: screenHeight,
        isForward: true,
        parent: _controller.view,
      );
    } else {
      return Tween<double>(
        begin: _kCartHeight,
        end: screenHeight,
      ).animate(
        CurvedAnimation(
          parent: _controller.view,
          curve: Interval(0.434, 1.0, curve: Curves.linear), // not used
          // only the reverseCurve will be used
          reverseCurve:
              Interval(0.434, 1.0, curve: Curves.fastOutSlowIn.flipped),
        ),
      );
    }
  }

  Animation<double> _getShapeAnimation() {
    if (_controller.status == AnimationStatus.forward) {
      return Tween<double>(begin: _kCornerRadius, end: 0.0).animate(
        CurvedAnimation(
          parent: _controller.view,
          curve: Interval(0.0, 0.3, curve: Curves.fastOutSlowIn),
        ),
      );
    } else {
      return _getEmphasizedEasingAnimation(
        begin: _kCornerRadius,
        peak: _getPeakPoint(begin: _kCornerRadius, end: 0.0),
        end: 0.0,
        isForward: false,
        parent: _controller.view,
      );
    }
  }

  Animation<double> _getThumbnailOpacityAnimation() {
    return Tween<double>(begin: 1.0, end: 0.0).animate(
      CurvedAnimation(
        parent: _controller.view,
        curve: _controller.status == AnimationStatus.forward
            ? Interval(0.0, 0.3)
            : Interval(0.532, 0.766),
      ),
    );
  }

  Animation<double> _getSheetOpacityAnimation() {
    return CurvedAnimation(
      parent: _controller.view,
      curve: _controller.status == AnimationStatus.forward
          ? Interval(0.3, 0.6)
          : Interval(0.766, 1.0),
    );
  }

  bool get _isOpen {
    final AnimationStatus status = _controller.status;
    return status == AnimationStatus.completed ||
        status == AnimationStatus.forward;
  }

  void open() {
    if (!_isOpen) {
      _controller.forward();
    }
  }

  void close() {
    if (_isOpen) {
      _controller.reverse();
    }
  }

  bool get _sheetIsVisible => _thumbnailOpacityAnimation.value == 0.0;

  Widget _buildThumbnails() {
    return ExcludeSemantics(
      child: Opacity(
        opacity: _thumbnailOpacityAnimation.value,
        child: Container(
          width: 70,
          height: 70,
          child: widget.sheetThumbnail,
        ),
      ),
    );
  }

  Widget _buildSheetScreen() {
    return Opacity(
      opacity: _cartOpacityAnimation.value,
      child: widget.pickerScreen,
    );
  }

  Widget _buildSheet(BuildContext context, Widget child) {
    final Size screenSize = MediaQuery.of(context).size;
    final double screenWidth = screenSize.width;
    final double screenHeight = screenSize.height;

    _widthAnimation = _getWidthAnimation(screenWidth);
    _heightAnimation = _getHeightAnimation(screenHeight);
    _shapeAnimation = _getShapeAnimation();
    _thumbnailOpacityAnimation = _getThumbnailOpacityAnimation();
    _cartOpacityAnimation = _getSheetOpacityAnimation();

    return Container(
      width: _widthAnimation.value + 20,
      height: _heightAnimation.value + 10,
      child: Material(
        animationDuration: Duration(milliseconds: 0),
        shape: RoundedRectangleBorder(
          borderRadius: BorderRadius.only(
            topLeft: Radius.circular(_shapeAnimation.value),
          ),
        ),
        elevation: 4.0,
        color: Colors.deepPurple,
        child: _sheetIsVisible ? _buildSheetScreen() : _buildThumbnails(),
      ),
    );
  }

  Widget _buildSlideAnimation(BuildContext context, Widget child) {
    _slideAnimation = _getEmphasizedEasingAnimation(
      begin: Offset(1.0, 0.0),
      peak: Offset(_kPeakVelocityProgress, 0.0),
      end: Offset(0.0, 0.0),
      isForward: widget.hideController.status == AnimationStatus.forward,
      parent: widget.hideController,
    );

    return SlideTransition(
      position: _slideAnimation,
      child: child,
    );
  }

  Future<bool> _onWillPop() async {
    if (!_isOpen) {
      await SystemNavigator.pop();
      return true;
    }

    close();
    return true;
  }

  @override
  Widget build(BuildContext context) {
    return AnimatedSize(
      key: _expandingPickerSheetKey,
      duration: Duration(milliseconds: 225),
      curve: Curves.easeInOut,
      vsync: this,
      alignment: FractionalOffset.topLeft,
      child: WillPopScope(
        onWillPop: _onWillPop,
        child: AnimatedBuilder(
          animation: widget.hideController,
          builder: _buildSlideAnimation,
          child: GestureDetector(
            behavior: HitTestBehavior.opaque,
            onTap: open,
            child: AnimatedBuilder(
              builder: _buildSheet,
              animation: _controller,
            ),
          ),
        ),
      ),
    );
  }
}